/*
 *  
 *  Copyright (c) 2004 Steve Slaven, All Rights Reserved.
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA 02111-1307 USA
 *  
*/
// This includes the map data handler, also handles caching and stuff like that
package net.hoopajoo.cheeselab3;
import java.io.*;
import java.util.*;
import java.awt.geom.Area;
import java.awt.Rectangle;
import java.awt.Point;
import java.awt.Dimension;
import jdbm.RecordManager;
import jdbm.RecordManagerFactory;
import jdbm.RecordManagerOptions;
import jdbm.htree.HTree;

class CheeseMapData implements CheeseMapListener {
    // Store and load operations on normal longs are not guaranteed to be
    // atomic thus everything that will be used for inter-thread communications
    // must be declared with the volatile keyword.
    // Actually our current implementation uses only last_map from multiple
    // threads so others can be declared as normal longs, but we'll be careful.
    private volatile long last_map = 0,
                          changed_map = 1,
                          last_request_map = 0,
                          request_map = 0;

    // These store the window for the last missed map request, so we can
    // re-run it later after the map has settled to make sure the
    // current view is up-to-date
    private Point3d request_start = new Point3d(),
                    request_end = new Point3d();

    private int memcache_size = 10000; // FIXME: Make this configurable
    private RecordManager db = null;
    private HTree cvars = null; // Internal persistent info about the cache (like last_sync)
    private Map index = null; // Room index
    private Map memcoords = null; // Coordinate index

    private HashMap players = new HashMap();
    //private Map memrooms = Collections.synchronizedMap(new SoftHashMap( memcache_size ) ); // faster memory access
    private Map memrooms = Collections.synchronizedMap(new HashMap( memcache_size ) ); // faster memory access
    private HashMap lastrequest = new HashMap(); // last mtime we requested by roomid so
    // we don't make redundant requests
    private boolean use_memcache = false;

    // This is the highest seen mtime, not the last saved one
    private long highest_mtime = 0;

    private Vector remap_events = new Vector();

    // For caching remap requests
    private HashMap mapped_areas = new HashMap();

    public CheeseMapDataStats stats = null;

    // We save remap requests in a vector so we can clear out the map area
    // whenever we receive a map event, this makes it so that deleted rooms
    // will disappear
    class RemappedWindow {
        Point3d start = new Point3d(),
                end = new Point3d();
    }

    CheeseMapData() {
        CheeseLab.dmesg( 2, "Creating map data instance" );
        CheeseMap.addCheeseMapListener( this );
    }

    // this is the last saved highest mtime
    public long get_highest_mtime() {
        return( highest_mtime );
    }

    public void initcache() {
        // Clear memcache
        if( memrooms != null )
            memrooms.clear();
        if( memcoords != null)
            memcoords.clear();

        opencache( true );

        use_memcache = CheeseMap.prop_get_boolean( "mapdata_memcache", true );
        CheeseLab.dmesg( 2, "Use memcache for map data: " + use_memcache );

        // get highest mtime from hash
        long max = Long.MAX_VALUE; // in case of exception do not refresh
        try {
            Long tmp_mtime = (Long)cvars.get( "highest_mtime" );
            if( tmp_mtime != null ) {
                CheeseLab.dmesg( 2, "Got last highest mtime: " + tmp_mtime );
                max = tmp_mtime.longValue();
            }else{
                max = 0;
            }
            // Load stats
            stats = (CheeseMapDataStats)cvars.get( "stats" );
            if( stats == null )
                stats = new CheeseMapDataStats();

        }catch( Exception e ) {
            e.printStackTrace();
        }
        highest_mtime = max;

        /* Construct coodrinate index, fill RecordManager Cache & store
         * rooms into memcache when preload is requested.
         */
        if( CheeseMap.prop_get_boolean( "mapdata_memcache_preload" ) ) {
            try {
                CheeseLab.dmesg( 2, "Preloading room cache in to memory" );
                Iterator keys = index.keySet().iterator();
                while( keys.hasNext() ) {
                    Integer key = (Integer)keys.next();
                    CheeseRoomData room = fetch_room( key );
                    memrooms.put( key, room );
                    put_roomid_by_coord( key, room.x, room.y, room.z );
                }
            }catch( Exception e ) {
                e.printStackTrace();
            }
        }
    }

    public void opencache( boolean trans ) {
        /* Close old room cache if it exists.
         * This will make sure it gets saved to disk properly.
         */
        if( db != null )
            closecache();

        try {
            File cache = new File( CheeseMap.cachedir + "/tiles" );
            cache.mkdirs();

            Properties o = new Properties();
            o.put( RecordManagerOptions.CACHE_SIZE, "2000" );

            if( !trans )
                o.setProperty( RecordManagerOptions.DISABLE_TRANSACTIONS, "true" );

            db = RecordManagerFactory.createRecordManager(
                     CheeseMap.cachedir + "/rooms.cache", o );

            long recid;

            recid = db.getNamedObject( "index" );
            if ( recid != 0 ) {
                index = (Map)db.fetch( recid );
            } else {
                index = Collections.synchronizedMap( new HashMap() ); // Room index
                db.setNamedObject( "index", db.insert( index ) );
            }

            recid = db.getNamedObject( "cvars" );
            if ( recid != 0 ) {
                cvars = HTree.load( db, recid );
            } else {
                cvars = HTree.createInstance( db );
                db.setNamedObject( "cvars", cvars.getRecid() );
            }

            recid = db.getNamedObject( "coordindex" );
            if ( recid != 0 ) {
                // No need to load coordinate cache when preloading (it will be constructed)
                if( !CheeseMap.prop_get_boolean( "mapdata_memcache_preload" ) )
                    memcoords = (Map)db.fetch( recid );
                else
                    memcoords = Collections.synchronizedMap( new HashMap() );
            } else {
                memcoords = Collections.synchronizedMap( new HashMap() );
                db.setNamedObject( "coordindex", db.insert( index ) );
            }

        }catch( Exception e ) {
            e.printStackTrace();
            CheeseLab.die( this, 1, "Unable to create room cache" );
        }
    }

    synchronized public void closecache() {
        try {

            if( db != null ) {
                commitcache();
                db.close();
            }

            index = null;
            cvars = null;
        } catch( Exception e ) {
            CheeseLab.stacktrace();
            CheeseLab.die( this, 1, "Unable to close room cache: " + e.toString() );
        }

        db = null;
        use_memcache = false;
    }

    // FIXME: Investigate synchronization issues, I'm sure open/close/commit
    //        aren't as thread safe as they should be.
    private void commitcache() {
        if( db == null )
            return;

        try {
            db.update( db.getNamedObject("index"), index );
            db.update( db.getNamedObject("coordindex"), memcoords );
            cvars.put( "stats", stats );
            CheeseLab.dmesg( 2, "Saving highest_mtime: " + highest_mtime );
            cvars.put( "highest_mtime", new Long( highest_mtime ) );
            db.commit();
        } catch (IOException e) {
            CheeseLab.stacktrace();
            CheeseLab.die( this, 1, "Unable to save room cache: " + e.toString() );
        }
    }

    synchronized private CheeseRoomData fetch_room( Integer roomid ) throws IOException {
        Long ie = (Long) index.get( roomid );
        CheeseRoomData room = null;

        if( ie != null )
            room = (CheeseRoomData)db.fetch( ie.longValue() );

        return room;
    }

    // Update a region of the map (most likely because we
    // are displaying it)
    public void remap( int sx, int sy, int sz, int ex, int ey, int ez ) {
        if( CheeseMap.prop_get_boolean( "sync_on_world", true ) ) {
            // Do not use remap, using sync-on-world logic
            return;
        }


        if( last_map == changed_map ) {
            // TODO: we have not finished our last map update (which will inc last_map)
            // so discard this request (but note it so that a thread
            // can clean up this loose-end later)
            request_start.x = sx;
            request_start.y = sy;
            request_start.z = sz;
            request_end.x = ex;
            request_end.y = ey;
            request_end.z = ez;
            request_map++;
            CheeseLab.dmesg( 1, "Skipping overactive remap, queuing for later: " + request_map );
            return;
        }

        changed_map = last_map;

        /* Check if we have already requested this area from the server.
         * This cache will be invalidated by world events.
         */
        boolean found = true;
        for( int z = sz; z <= ez; z++ ) {
            Area area51 = (Area)mapped_areas.get( new Integer(z) );
            if( area51 == null ||
                    !area51.contains( sx, sy, ex - sx, ey - sy ) ) {
                found = false;
                break;
            }
        }

        if( found ) {
            CheeseLab.dmesg( 3, "Area already cached, map request not sent" );
            last_map++;
            return;
        }
        // TODO: Notify canvas that there is no point to wait

        /* Now once we are already doing the round-trip, better ask for
         * a bit bigger chunk so that we will not need to do it again too
         * soon.
         * TODO: Make this pre-fetching a configurable option?
         */
        int pfx = 10, pfy = pfx, pfz = 1;
        sx -= pfx; ex += pfx; sy -= pfy; ey += pfy; sz -= pfz; ez += pfz;
        if( sx < 0 ) sx = 0; if( sy < 0 ) sy = 0; if( sz < 0 ) sz = 0;

        RemappedWindow w = new RemappedWindow();
        w.start.x = sx;
        w.start.y = sy;
        w.start.z = sz;
        w.end.x = ex;
        w.end.y = ey;
        w.end.z = ez;

        remap_events.add( w );

        CheeseMap.send_tx( 1, "map",
                           sx + "," + sy + "," + sz + " " +
                           ex + "," + ey + "," + ez );
    }

    // This should be periodically called by a thread to make sure we
    // get caught up to the latest view if we had to skip out on some remap
    // requests
    public void catchup_remap() {
        if( request_map != last_request_map ) {
            CheeseLab.dmesg( 1, "Trying to catch up remap requests: " + last_request_map );
            last_request_map = request_map;
            remap( request_start.x, request_start.y, request_start.z,
                   request_end.x, request_end.y, request_end.z );
        }
    }

    // This just queues room updates for stuff in our cache that is out
    // of date, this way we only get updates for what is in our
    // viewable area but we also can use old cached data in the mean time
    public void update_map( byte[] buffer ) {
        String s = new String( buffer );
        BufferedReader in = new BufferedReader( new StringReader( s ) );
        String line;

        if( index == null ) // Means Sync Cache operation is in progress
            return;

        // Get last remap req (if applicable)
        RemappedWindow w = (RemappedWindow)remap_events.get( 0 );

        boolean [][][] clean_keys = new boolean[w.end.z - w.start.z + 1]
                                    [w.end.x - w.start.x + 1]
                                    [w.end.y - w.start.y + 1];
        if( w != null ) {
            remap_events.remove( 0 );

            // We make a vector of items to null-out based on all the keys
            // that could be set from this remap event.
            /* Actually using hashmap or vector for this purpose is the slowest
             * possible option imaginable, addition and deletion from hashmaps are
             * sloooooow. Using a boolean array instead.
             */
            for( int x = w.start.x; x <= w.end.x; x++ ) {
                for( int y = w.start.y; y <= w.end.y; y++ ) {
                    for( int z = w.start.z; z <= w.end.z; z++ ) {
                        clean_keys[z - w.start.z][x - w.start.x][y - w.start.y] = true;
                    }
                }
            }
        }

        // Process incoming data
        try{
            while( (line = in.readLine()) != null ) {
                String[] values = line.split( "," );
                int roomid = Integer.parseInt( values[ 0 ] );
                int x = Integer.parseInt( values[ 1 ] );
                int y = Integer.parseInt( values[ 2 ] );
                int z = Integer.parseInt( values[ 3 ] );

                // Simply queue a request for the latest room data here,
                // if we already have something then it will be cached and
                // the renderer will do Do the Right Thing (tm) and render
                // what we have, and new room events will be generated as
                // updates come in and it'll repaint those
                CheeseRoomData room = get_room( roomid, Long.parseLong( values[ 4 ] ) );

                // And poke what's to come in to our x/y/z array (since the final room
                // update will simply do it again, it's harmless but allows the renderer
                // to draw the ?? boxes where rooms will fill in to) and the room update
                // handler does the right thing to remove old reverse entries.  The only
                // drawback here is it's possible that a room may exist in 2 places on the
                // map at once until a room update is received
                //
                // TODO: It might be nice to automagically remove the old room someday though, maybe
                // if we had a common update_coords_cache() function that could be called here
                // or in the update_room subs that would be better?
                
                try {
                    clean_keys[z - w.start.z][x - w.start.x][y - w.start.y] = false;
                    
                    Integer val = new Integer( roomid );
                    put_roomid_by_coord( val, x, y, z );

                }catch( Exception e ) {
                    e.printStackTrace();
                }
            }

            // Any keys left in clean_keys are keys that need to be unset in the coords
            // and memcoords hashes
            for( int x = w.start.x; x <= w.end.x; x++ ) {
                for( int y = w.start.y; y <= w.end.y; y++ ) {
                    for( int z = w.start.z; z <= w.end.z; z++ ) {
                        if( clean_keys[z - w.start.z][x - w.start.x][y - w.start.y] ) {
                            String coordkey = x + "," + y + "," + z;
                            memcoords.put( coordkey, new Integer[] { new Integer( -1 ) } );
                        }
                    }
                }
            }            
        }catch( Exception e ){
            e.printStackTrace();
        }
        last_map++;
        
    }

    public void update_world() {
        CheeseLab.dmesg( 3, "World changed! Killing area cache." );

        mapped_areas.clear();
    }

    /**
     * Save room to cache.
     * @param roomid room id
     * @param room room data
     */
    synchronized private void save_room(Integer roomid, CheeseRoomData room) throws IOException {
        Long ie = (Long)index.get( roomid );

        if( ie == null ) {
            ie = new Long( db.insert( room ) );
            index.put( roomid, ie );
        }
        else {
            db.update( ie.longValue(), room );
        }
    }

    synchronized private void delete_room(Integer roomid) throws IOException {
        Long ie = (Long)index.get( roomid );

        if( ie != null ) {
            db.delete( ie.longValue() );
        }
        index.remove( roomid );
    }

     public CheesePlayerData[] get_players() {
        Vector data = new Vector();
        try {
            //JDBMEnumeration values = players.values();
            Iterator values = players.values().iterator();
            while( values.hasNext() )
                data.add( values.next() );

            //while( values.hasMoreElements() )
            //	data.add( values.nextElement() );
        }catch( Exception e ) {
            CheeseLab.dmesg( 0, "Exception in get_players: " + e );
            e.printStackTrace();
        }
        return( (CheesePlayerData[])data.toArray( new CheesePlayerData[0] ) );
    }

    public CheesePlayerData get_player( String name ) {
        try {
            return( (CheesePlayerData)players.get( name ) );
        }catch( Exception e ) {
            return( null );
        }
    }
    
    /* Oh jeez, we keep roomids in an array simply because in theory
     * You shouldn't have two rooms at the same spot on Your map, but
     * when You are editing it and move rooms around they sometimes
     * end up being on top of each other, then when You move those
     * rooms again, the old one that used to be there just disappears
     * from the map (it's still on server and room cache but it's not shown)
     */
    public Integer get_roomid_by_coord( int x, int y, int z ) {
        Integer [] roomids = null;
        String coordkey = x + "," + y + "," + z;
        
        roomids = (Integer [])memcoords.get( coordkey );
        
        if ( roomids == null )
            return null;
        else // If multiple rooms, return only the first one
            return roomids[ 0 ];
    }
    
    /**
     * Insert roomid to coordinate cache
     */
    private void put_roomid_by_coord( Integer roomid, int x, int y, int z ) {
        String coordkey = x + "," + y + "," + z;
        Integer [] old_roomids = (Integer[]) memcoords.get( coordkey );

        if( old_roomids == null || old_roomids[ 0 ].intValue() == -1 ) {
            memcoords.put( coordkey, new Integer[] { roomid } );
            return;
        }
        
        /* Search for the roomid, of course we could sort the array and
         * use a much faster binary search instead, but those arrays are
         * going to be small anyway and doing it this way ensures that
         * most recent changes will always be on top and will not confuse
         * the user when rooms are being moved around.
         */
        int i;
        for( i = 0; i < old_roomids.length; i++ ) {
            if( old_roomids[ i ].equals( roomid ) )
                break;
        }
        
        if( i < old_roomids.length ) {
            return; // Do nothing, it's already there
        }else{
            Integer [] new_roomids = new Integer[old_roomids.length + 1];
            System.arraycopy(old_roomids, 0, new_roomids, 1, old_roomids.length );
            new_roomids[ 0 ] = roomid;
            memcoords.put( coordkey, new_roomids );
        }
    }
    
    /**
     *  Clear specified roomid from coordinate cache position
     */
    private void del_roomid_by_coord( Integer roomid, int x, int y, int z ) {
        String coordkey = x + "," + y + "," + z;
        
        Integer [] old_roomids = (Integer [])memcoords.get( coordkey );
        
        // If no entry or has only one room and it matches set to -1
        if( old_roomids == null ||
            (old_roomids.length == 1 && old_roomids[0].equals( roomid ) ) ) {
            memcoords.put( coordkey, new Integer[] { new Integer( -1 ) } );
        }else {
            int i;
            for( i = 0; i < old_roomids.length; i++ ) {
                if( old_roomids[ i ].equals( roomid ) )
                    break;
            }
            
            /* If found, cut it out ! */
            if( i < old_roomids.length ) {
                Integer [] new_roomids = new Integer[old_roomids.length - 1];
                System.arraycopy( old_roomids, 0, new_roomids, 0, i);
                System.arraycopy( old_roomids, i + 1,
                        new_roomids, i, old_roomids.length - i - 1);
                memcoords.put( coordkey, new_roomids );
            }
        }
        // roomid was not found, do nothing *blink*
    }

    // Get the room at coords x/y/z, may return NULL
    public CheeseRoomData get_room( int x, int y, int z ) {
        Integer roomid = null;
        

        // FIXME: Race condition! Some threads might access get_room before the cache
        //        is initialized
        if( memcoords == null )
            return null;

        roomid = get_roomid_by_coord( x, y, z );
        
        // A roomid of -1 in the memcoords means it was not a cache miss but
        // there is no room there
        // so we make it null here (otherwise missing rooms result in
        // cache misses and slow down rendering)
        if( roomid != null && roomid.intValue() == -1 )
            roomid = null;

        if( roomid == null ) {
            return( null );
        }else{
            return( get_room( roomid.intValue() ) );
        }
    }

    // Give mtime as the servers latest mtime, if our room is
    // at least that new we will not even bother with a round-trip
    public CheeseRoomData get_room( int id ) {
        return( get_room( id, 1 ) );
    }

    public CheeseRoomData get_room( int id, long mtime ) {
        return get_room( id, mtime, true );
    }

    public CheeseRoomData get_room( int id, long mtime, boolean askserver ) {
        if( use_memcache )
            return( mem_get_room( id, mtime, askserver ) );

        // Skip memcache code
        return( disk_get_room( id, mtime, askserver ) );
    }

    private CheeseRoomData mem_get_room( int id, long mtime, boolean askserver ) {
        // Get rooms from memcache first, if not there then
        // ask the disk cache (which in turn asks the server
        // if it fails)
        CheeseRoomData room = (CheeseRoomData)memrooms.get( new Integer(id) );
        if( room != null )
            if( mtime > room.mtime )
                room = null;

        if( room == null ) {
            room = disk_get_room( id, mtime, askserver );
            /* We really should implement some sort of 
             * random early drop policy for our roomcache.
             * The idea is that we delete a bunch of random rooms from cache
             * before it gets too full and therefore hopefully save
             * us the trouble and overhead of implementing an MRU (most
             * recently used) policy.
             */
            if( memrooms.size() >= memcache_size * 0.95 ) {
                // Creating an iterator is expensive, destroy 1/5 of the cache
                Iterator it = memrooms.keySet().iterator();
                int n = memcache_size / 5;
                Object [] keys = new Object[n];
                try {
                    for(int i = 0; i < n; i++ ) {
                        keys[i] = it.next();
                    }
                }catch(ConcurrentModificationException e) { }
                for( int i = 0; i < n; i++ ) {
                    memrooms.remove( keys[i] );
                }
            }
            
            memrooms.put( new Integer( id ), room );
        }
        return( room );
    }

    private CheeseRoomData disk_get_room( int id, long mtime, boolean askserver ) {
        // These are not valid rooms
        if( id < 1 ) {
            CheeseLab.dmesg( 1, "Invalid roomid requested: " + id );
            CheeseLab.stacktrace();
            CheeseRoomData room = new CheeseRoomData();
            room.name = "Invalid room";
            room.terrain = "?";
            room.overlays = new String[ 0 ];
            room.mtime = 0;
            return( room );
        }

        try {
            CheeseRoomData room = fetch_room( new Integer(id) );

            if( room == null && askserver ) {
                // Send back a dummy room
                room = new CheeseRoomData();
                room.name = "Uncached Room (loading)";
                room.terrain = "?";
                room.overlays = new String[0];
                room.mtime = 0;
            }

            if( askserver && room.mtime < mtime ) {
                Long last_requesttime = (Long)lastrequest.get( new Integer( id ) );
                if( last_requesttime == null || last_requesttime.longValue() < mtime ) {
                    CheeseLab.dmesg( 2, "Requesting room " + room.roomid + " (" + room.mtime + "<" + mtime + ")" );
                    // Request it from the server, a refresh should
                    // be issued when it comes in sooo :p
                    // you ought to get a "filling in" effect while spamming about
                    CheeseMap.send_tx( 1, "get", id + ":" + room.mtime );
                    lastrequest.put( new Integer( id ), new Long( mtime ) );
                }
            }

            return( room );
        }catch( Exception e ) {
            return( null );
        }
    }

    /* CheeseMapListener interface functions */

    public void roomChanged(CheeseRoomData room, CheeseRoomData oldroom) {
        CheeseLab.dmesg( 2, "Updating room " + room.roomid + " in cache" );

        Integer roomid = new Integer( room.roomid );

        try {
            // Save basic room data
            save_room( roomid, room );

            if( room.mtime > highest_mtime ) {
                highest_mtime = room.mtime;
            }

            /* Update coordinate cache */
            if( oldroom != null ) {
                del_roomid_by_coord( new Integer( oldroom.roomid ),
                        oldroom.x, oldroom.y, oldroom.z );
            }

            put_roomid_by_coord( roomid, room.x, room.y, room.z );

            /* Update memory cache */
            if( use_memcache ) {
                memrooms.put( roomid, room );
            }
        }catch( Exception e ) {
            CheeseLab.dmesg( 0, "Error saving object room: " + e );
            e.printStackTrace();
        }

        // Update statistics
        if( stats.max_x < room.x ) stats.max_x = room.x;
        if( stats.max_y < room.y ) stats.max_y = room.y;
        if( stats.max_z < room.z ) stats.max_z = room.z;
        if( stats.min_x > room.x ) stats.min_x = room.x;
        if( stats.min_y > room.y ) stats.min_y = room.y;
        if( stats.min_z > room.z ) stats.min_z = room.z;
        if( oldroom == null ) { // New room
            stats.total_rooms++;
            int n = 0;
            Integer zkey = new Integer( room.z );
            Integer nr = (Integer)stats.nrooms_z.get( zkey );
            if( nr != null )
                n = nr.intValue();
            n++;
            stats.nrooms_z.put( zkey, new Integer( n ) );
        }
    }

    public void roomDeleted(CheeseRoomData room) {
        CheeseLab.dmesg( 2, "Deleting room " + room.roomid + " from cache" );

        Integer roomid = new Integer( room.roomid );

        // Delete from coordinate cache
        del_roomid_by_coord( roomid, room.x, room.y, room.z );

        if( use_memcache ) {
            memrooms.remove( roomid );
        }

        // Delete from persistent cache
        try {
            delete_room( roomid );
        } catch (IOException e) {
            CheeseLab.dmesg( 0, "Error deleting room from cache: " + e );
            e.printStackTrace();
        }

        // TODO: Update statistics. Unfortunately this is not very easy to do :(
    }

    public void onlineChanged(boolean online) {
        if( online )
            initcache();
        else
            closecache();
    }

    public void playerChanged(CheesePlayerData player, CheesePlayerData oldplayer) {
        CheeseLab.dmesg( 2, "Updating player" );

        try {
            CheeseLab.dmesg( 3, "Saving player: " + player.name );
            players.put( player.name, player );
        }catch( Exception e ) {
            CheeseLab.dmesg( 0, "Error saving player object player: " + e );
            e.printStackTrace();
        }
    }

    public void tileChanged(String fname) { }
    public void refreshStarted() { }

    public void refreshFinished() {
        commitcache();
    }
    
    public String get_sysinfo() {
        String sysinfo = "";
        
        if( stats != null ) {
            sysinfo += "Total rooms on map: " + stats.total_rooms + "\n";
            /*sysinfo += "Min coordinates: " + stats.min_x
                    + "," + stats.min_y + "," + stats.min_z
                    + "\n" + "Max coordinates: " + stats.max_x
                    + "," + stats.max_y + "," + stats.max_z
                    + "\n";*/
        }
        
        if( memrooms != null )
            sysinfo += "Rooms in memcache: " + memrooms.size() + "\n";

        return sysinfo;
    }
}
